@@ -1,7 +1,5 @@
-/* $Xorg: NextEvent.c,v 1.8 2001/02/09 02:03:55 xorgcvs Exp $ */
-
 /***********************************************************
-Copyright 1993 Sun Microsystems, Inc.  All rights reserved.
+Copyright (c) 1993, Oracle and/or its affiliates. All rights reserved.
 
 Permission is hereby granted, free of charge, to any person obtaining a
 copy of this software and associated documentation files (the "Software"),
@@ -69,7 +67,6 @@ used in advertising or otherwise to promote the sale, use or other dealings
 in this Software without prior written authorization from The Open Group.
 
 */
-/* $XFree86: xc/lib/Xt/NextEvent.c,v 3.26 2002/06/04 21:55:42 dawes Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -359,7 +356,21 @@ static int IoWait (
     return Select (wf->nfds, &wf->rmask, &wf->wmask, &wf->emask,
 		   wt->wait_time_ptr);
 #else
-    return poll (wf->fdlist, wf->fdlistlen, wt->poll_wait);
+    int ret = poll (wf->fdlist, wf->fdlistlen, wt->poll_wait);
+    /* If poll() returns an event we didn't expect, such as POLLNVAL, treat
+     * it as if it failed. */
+    if(ret >= 0) {
+        nfds_t i;
+        for (i=0; i < wf->fdlistlen; i++) {
+            struct pollfd *fd = &wf->fdlist[i];
+            if (fd->revents & ~fd->events) {
+                ret = -1;
+                errno = EIO;
+                break;
+            }
+        }
+    }
+    return ret;
 #endif
 }
 
@@ -734,7 +745,14 @@ WaitLoop:
 #endif
 	return dpy_no;
     }
-    goto WaitLoop;
+    if (block)
+        goto WaitLoop;
+    else {
+#ifdef USE_POLL
+	XtStackFree ((XtPointer) wf.fdlist, fdlist);
+#endif
+        return -1;
+    }
 }
 
 #define IeCallProc(ptr) \
